/***********************************************************************\
* NLED Text Editor Copyright 1987-2003, Evan Alexander Weaver           *
*                                                                       *
* This source code, and the associated executable programs obtained by  *
* compiling it, are published under the GNU General Public License. See *
* the accompanying file "nled.txt" for the details of this license and  *
* some other comments about NLED and this source code.                  *
*                                                                       *
* File: NLEDISP.C                         Date Last Modified: 29-Apr-03 *
*                                                                       *
* Various functions controlling NLED's screen display.                  *
* Functions are:                                                        *
*   dispcalc() - recalculates scr_lxxx variables from a given row       *
*   dispinit() - initilializes scr_lxxx for a new file                  *
*   display() - displays that part of the screen given by draw_xxx vars *
*   dispmess() - displays null terminated command line message          *
*   dispnmess() - displays fixed length command line message            *
*   dispnextw() - moves cursor to start of next word                    *
*   dispprevw() - moves cursor to start of previous word                *
*   dispstatus() - displays status line                                 *
*   dispupd() - processes cursor movement requests (updates display)    *
*   setdraw() - sets draw_xxx vars to draw at least current line        *
*   setdrawall() - sets draw_xxx vars to draw whole screen              *
*                                                                       *
* 05-Oct-91: ANSI function headers instead of "old-style"               *
* 12-Feb-92: UNIX standardization                                       *
* 20-Mar-95: Linux port                                                 *
* 09-Jul-96: Win95 port                                                 *
* 03-Feb-02: 2.5 cleanup release                                        *
* 15-Feb-02: 2.51 release                                               *
* 29-Apr-03: 2.52 release (screen buffer fix for Windows only)          *
\***********************************************************************/

#include <stdio.h>
#include <string.h>
#include "nled.h"
#include "nledkeys.h"
#include "nledvdef.h"
#include "nledfdef.h"

/***********************************************************************\
* Calculates the scr_lxxx variables from row 'row' to the end, assuming *
* that theoffset (scr_loff) is correct for 'row'. Also recalculates     *
* abs_pos.                                                              *
\***********************************************************************/
void dispcalc(int row)
{
    INDEX offset;

    offset = scr_loff[row];
    while ((row < num_rows-1) && (offset < len_ds)) {
        scr_loff[row] = offset;
        offset += 1 + (scr_llen[row] = datllen(offset));
        row++;
    }
    while (row < num_rows-1) {
        scr_loff[row]=len_ds;
        scr_llen[row]=0;
        row++;
    }
    abs_pos = scr_loff[row_no] + ((scr_left + col_no > scr_llen[row_no]) ?
     scr_llen[row_no] : scr_left + col_no);
}

/**********************************************************************\
* Displays a screen-full of the text file being edited, starting at    *
* character # 'charac' on line # 'line', while initializing all        *
* related global display control variables. ('rowno','colno') gives    *
* the position on the screen for the cursor to be placed. The function *
* searches for the correct starting place from the beginning of the    *
* file data string.                                                    *
\**********************************************************************/
void dispinit(INDEX line, INDEX charac, int rowno, int colno)
{
    INDEX i,lctr;

    qd_move(0,0);
    scr_top = line;
    scr_left = charac;
    row_no = rowno;
    col_no = colno;

    i=0;
    for (lctr = 0; (lctr < scr_top) && (i < len_ds); lctr++) {
        i += (1 + datllen(i));
    }

    /* we've found the first character on the first line */
    if (lctr == scr_top) {
        scr_loff[0] = i;
        dispcalc(0);
    }
    else { /* we've been asked to display past the end of file */
        scr_top = lctr;
        scr_loff[0] = i;
        dispcalc(0);
    }
    setdrawall();
    block_beg = block_end = 0;
}

/**********************************************************************\
* Redraws the part of the screen specified by the 'draw_***' globals.  *
* All the proper global settings except abs_pos must be already set.   *
* 01-APR-87: changed "startc" & "endc" to int                          *
* 23-APR-87: added calls to qd_off and qd_on                           *
* 27-APR-87: don't blank screen if only one line is being updated      *
* 09-Jul-96: turn cursor off even if only redrawing one line           *
\**********************************************************************/
void display(void)
{
    INDEX starti;
    int endc,len,n,r,startc;

    if (draw_brow > num_rows - 2) return;

    qd_off();
    startc = draw_col;
    for (r = draw_brow; r <= draw_erow; r++) {
        qd_move(r,startc);
        if (scr_llen[r] <= scr_left)
            endc = 0;
        else
            endc = min(num_cols, scr_llen[r] - scr_left);
        len = endc - startc;
        starti = scr_loff[r] + scr_left + startc;
        if (len > 0) {
            if (starti < block_beg) {
                n = min(len, block_beg - starti);
                datprint(starti,n);
                starti += n;
                len -= n;
            }
            if (len > 0) {
                if (starti < block_end) {
                    qd_boldon();
                    n = min(len, block_end - starti);
                    datprint(starti,n);
                    starti += n;
                    len -= n;
                    qd_boldoff();
                }
                if (len > 0)
                    datprint(starti,len);
            }
        }
        if (endc < num_cols) qd_eeol();
        startc = 0;
    }
    qd_on();
    draw_brow = num_rows - 1;
    draw_erow = -1;
    draw_col = num_cols;
}

/***********************************************************************\
* Displays null-terminated 'string[]' on the status line and places the *
* cursor at column 'col' on the status line (if >= 0). 'urgent' is TRUE *
* or FALSE, and determines the urgency of the need to display the       *
* message (see dispnmess() for more information).                       *
\***********************************************************************/
void dispmess(char *string, int col, int urgent)
{
    dispnmess(string, strlen(string), col, urgent);
}

/***********************************************************************\
* Displays 'len' characters, beginning at '*string', on the status line *
* starting at column 0. The cursor is positioned at column 'col' on the *
* status line (if positive) or is left in the text part of the screen   *
* (if 'col' is -1). If 'urgent' is FALSE, the message is not displayed  *
* until the console is opened for input (& may never appear if a        *
* function key is supplying input) - this is used for displaying prompt *
* messages; if it is TRUE, the message is displayed immediately - this  *
* is used to display ongoing operation status messages.                 *
\***********************************************************************/
void dispnmess(char *string, int len, int col, int urgent)
{
    int i;

    if (len > MAX_PM + 1) len = MAX_PM + 1;
    for (i = 0; i < len; i++)
        prompt_msg[i] = string[i];
    while (i <= MAX_PM)
        prompt_msg[i++] = ' ';
    msg_col = col;
    msg_flag = TRUE;

    if (urgent) {
        dispstatus();
        msg_flag = FALSE;
    }
}

/***********************************************************************\
* Moves the cursor to the beginning of the next word, scrolling the     *
* screen if necessary. A word is defined as non-blank characters        *
* separated from others by one or more blanks.                          *
\***********************************************************************/
void dispnextw(void)
{
    INDEX dcol,drow,pos;
    int switches;

    dcol = 0;
    drow = 0;
    switches = 0;
    pos = abs_pos;

    while ((switches < 2) && (pos < len_ds)) {
        switch(datget(pos)) {
        case LF:
            if (0 == switches) switches++;
            dcol = 0;
            drow++;
            pos++;
            break;

        case ' ':
            if (0 == switches) switches++;
            dcol++;
            pos++;
            break;

        default:
            if (0 < switches) switches++;
            else {
                pos++;
                dcol++;
            }
            break;
        }
    }

    if (drow == 0) {
        dispupd('c',dcol);
    }
    else {
        dispupd('r',drow);
        dispupd('c',-col_no - scr_left + dcol);
    }
}

/***********************************************************************\
* Moves the cursor to the beginning of the previous word, scrolling the *
* screen if necessary. A word is defined as non-blank characters        *
* separated from others by one or more blanks.                          *
\***********************************************************************/
void dispprevw(void)
{
    INDEX dcol,drow,pos;
    int switches;

    dcol = 0;
    drow = 0;
    switches = 0;
    pos = abs_pos - 1;

    /* if we're past the end of the line, move back to the end */
    if (scr_left + col_no >= scr_llen[row_no])
        dispupd('c',scr_llen[row_no] - scr_left - col_no);

    while ((switches < 2) && (pos >= 0)) {
        switch(datget(pos)) {
        case LF:
            if (0 < switches) switches++;
            else {
                dcol = 0;
                drow--;
                pos--;
            }
            break;

        case ' ':
            if (0 < switches) switches++;
            else {
                dcol--;
                pos--;
            }
            break;

        default:
            if (0 == switches) switches++;
            dcol--;
            pos--;
            break;
        }
    }

    if (drow == 0)
        dispupd('c',dcol);
    else {
        dispupd('r',drow);
        dispupd('c',scr_llen[row_no] - col_no - scr_left + dcol);
    }
}

/***********************************************************************\
* Displays the status line.                                             *
* 01-APR-87: Updated version number to 1.1                              *
* 16-APR-87: Updated version number to 1.2                              *
* 10-Feb-88: Updated version number to 1.3                              *
* 22-Apr-88: Updated version number to 1.4                              *
* 30-Dec-88: Updated version number to 1.5,SNLED                        *
* 13-Mar-89: Updated version number to 1.6 beta                         *
* 05-Oct-91: Updated version number to 1.6                              *
* 05-Oct-91: Updated version number to 1.7 beta                         *
* 21-Feb-92: Updated version number to 2.0 (UNIX/DOS combined)          *
* 28-Mar-92: Updated version number to 2.1                              *
* 02-Sep-94: Updated version number to 2.3, UNIX status disable         *
* 03-Apr-95: Updated version number to 2.4 beta d (Linux)               *
* 09-Jul-96: Updated version number to 2.4 beta e (Win95)               *
* 03-Feb-02: Updated version number to 2.5, display last 20 chars of    *
*            file name, not first 22                                    *
* 15-Feb-02: Updated version number to 2.51                             *
* 29-Apr-03: Updated version number to 2.52 (Windows screen buffer fix) *
\***********************************************************************/
void dispstatus(void)
{
    int n;

    qd_off();
    qd_revon();
    qd_move(num_rows-1,0);
    if (msg_flag) {
        /* there is a message to display */
        printsstr(0,MAX_PM+1,prompt_msg);
    }
    else {
        /* there is no message, so use default status line format */
#ifdef SHAREWARE
        printstr(" SNLED2.52 File:");
#else
        printstr(" NLED 2.52 File:");
#endif
        n = strlen(file_name);
        if (n > 22) {
            printsstr(0, 2, "..");
            printsstr(n - 20, 20, file_name);
        }
        else
            printsstr(0,n,file_name);
        if (0 < (n = 22 - n))
            printsstr(0,n,"                      ");
#if UNIX
      if (status_disable)
        printstr("            ");
      else {
#endif
        printstr(" Pos:");
        printindex(1 + abs_pos,7);
#if UNIX
      }
#endif
    }
#if UNIX
  if (status_disable)
    printstr("                         ");
  else {
#endif
    printstr("Line:");
    printindex(1 + scr_top + row_no,7);
    printstr("Col:");
    printindex(1 + scr_left + col_no,7);
    printhex(datget(abs_pos));
#if UNIX 
  }
#endif
    /* don't print in last column! */
    printstr((ins_mode) ? " Ins": " Ovr");
    /* highlight the rest of the line */
    qd_eeol();
    qd_revoff();
    if (msg_flag && (msg_col >= 0))
        qd_move(num_rows-1,msg_col);
    else
        qd_move(row_no,col_no);
    qd_flush();
    qd_on();
}

/**********************************************************************\
* Moves the cursor position vertically or horizontally, scrolling the  *
* screen if necessary. 'delta' is the number of characters to move, &  *
* 'mode' is 'r' (move up or down Rows), 'c' (move left or right Cols), *
* 't' (scroll the Top of the screen up or down without moving cursor   *
* position on the display surface), or 'l' (scroll the Left column of  *
* the screen left or right without moving cursor on display surface).  *
* 01-APR-87: Fixed scrolling of screen >32767 in 'c' and 'r' modes     *
\**********************************************************************/
void dispupd(char mode, INDEX delta)
{
    INDEX i,j,maxscroll;
    int lctr;

    if (delta == 0) return;

    switch (mode) {
    case 't':
        if (delta > 0) { /* scroll screen up */
            maxscroll = nol_ds - scr_top - num_rows + 1;
            if (delta > maxscroll) delta = maxscroll;
            if (delta <= 0) return;
            if (delta < num_rows - 1) {
                for (lctr = 0; lctr < num_rows-1-delta; lctr++) {
                    scr_loff[lctr]=scr_loff[lctr+(int)delta];
                    scr_llen[lctr]=scr_llen[lctr+(int)delta];
                }
                i = 1 + scr_loff[lctr-1] + scr_llen[lctr-1];
            }
            else {
                j = num_rows - 1;
                i = 1 + scr_loff[num_rows - 2] + scr_llen[num_rows - 2];
                while (j < delta) {
                    i += 1 + datllen(i);
                    j++;
                }
                lctr = 0;
            }
            while ((lctr<num_rows-1) && (i<len_ds)) {
                scr_loff[lctr] = i;
                i += 1 + (scr_llen[lctr] = datllen(i));
                lctr++;
            }
            while (lctr<num_rows-1) {
                scr_loff[lctr] = len_ds;
                scr_llen[lctr] = 0;
                lctr++;
            }
            scr_top += delta;
            if (delta == 1) {
                qd_scrollup(0, num_rows - 2);
                if (draw_brow > 0) draw_brow--;
                draw_col = 0;
                draw_erow = num_rows - 2;
            }
            else
                setdrawall();
        }
        else { /* scroll screen down */
            delta = -delta;
            if (delta > scr_top) delta = scr_top;
            if (delta <= 0) return;
            if (delta < num_rows - 1) {
                for (lctr = num_rows - 2; lctr >= delta; lctr--) {
                    scr_loff[lctr] = scr_loff[lctr-(int)delta];
                    scr_llen[lctr] = scr_llen[lctr-(int)delta];
                }
                i = scr_loff[lctr+1];
            }
            else {
                j = 0;
                i = scr_loff[0];
                while (j < delta - num_rows + 1) {
                    i -= 1 + datpllen(i);
                    j++;
                }
                lctr = num_rows - 2;
            }
            while (lctr >= 0) {
                i -= 1 + (scr_llen[lctr] = datpllen(i));
                scr_loff[lctr] = i;
                lctr--;
            }
            scr_top -= delta;
            if (delta == 1) {
                qd_scrolldown(0, num_rows - 2);
                if (draw_erow < 0) draw_erow = 0;
                else if (draw_erow < num_rows - 2) draw_erow++;
                draw_col = 0;
                draw_brow = 0;
            }
            else
                setdrawall();
        }
        break;

    case 'l':
        scr_left += delta;
        if (scr_left < 0) scr_left = 0;
        draw_brow = 0;
        draw_erow = num_rows - 2;
        draw_col = 0;
        break;

    case 'r':
        if (delta > nol_ds - scr_top - row_no - 1)
            delta = nol_ds - scr_top - row_no - 1;
        if (delta < -scr_top - row_no)
            delta = -scr_top - row_no;
        delta += row_no;
        if (delta < 0) {
            dispupd('t',delta);
            row_no = 0;
        }
        else if (delta > num_rows - 2) {
            dispupd('t',delta - num_rows + 2);
            row_no = num_rows - 2;
        }
        else
            row_no = delta;
        break;

    case 'c':
        if (delta < -scr_left - col_no)
            delta = -scr_left;
        else
            delta += col_no;
        if (delta < 0) {
            dispupd('l',delta);
            col_no = 0;
        }
        else if (delta > num_cols - 1) {
            dispupd('l',delta - num_cols + 1);
            col_no = num_cols - 1;
        }
        else
            col_no = delta;
        break;

    default: /* no action used to force abs_pos recalculation */;
        break;
    }
    abs_pos = scr_loff[row_no] + ((scr_left + col_no > scr_llen[row_no]) ?
     scr_llen[row_no] : scr_left + col_no);
}

/***********************************************************************\
* Sets the draw_xxx variables to indicate redrawing of the screen from  *
* at least the current position to the end of the current line.         *
\***********************************************************************/
void setdraw(void)
{
     if (draw_brow > row_no) draw_brow = row_no;
     if (draw_erow < row_no) draw_erow = row_no;
     if (draw_col > col_no)  draw_col = col_no;
}

/***********************************************************************\
* Sets the draw_xxx variables to indicate redrawing of the whole screen *
\***********************************************************************/
void setdrawall(void)
{
    draw_brow = 0;
    draw_erow = num_rows-2;
    draw_col = 0;
}
/*--------------------------END-OF-NLEDISP.C---------------------------*/
